@@ -4,11 +4,11 @@
 use serde::de::DeserializeOwned;
 use serde::{Deserialize, Serialize};
 
-use cosmwasm_std::{from_slice, Binary, Order, Pair, StdError, StdResult, Storage};
+use cosmwasm_std::{from_slice, Binary, Order, Record, StdError, StdResult, Storage};
 
 use crate::helpers::namespaces_with_key;
-use crate::keys::EmptyPrefix;
 use crate::map::Map;
+use crate::prefix::{namespaced_prefix_range, PrefixBound};
 use crate::{Bound, Prefix, Prefixer, PrimaryKey, U32Key};
 
 pub fn index_string(data: &str) -> Vec<u8> {
@@ -60,6 +60,30 @@ where
     T: Serialize + DeserializeOwned + Clone,
 {
     // TODO: make this a const fn
+    /// Create a new MultiIndex
+    ///
+    /// idx_fn - lambda creating index key from value (first argument) and primary key (second argument)
+    /// pk_namespace - prefix for the primary key
+    /// idx_namespace - prefix for the index value
+    ///
+    /// ## Example:
+    ///
+    /// ```rust
+    /// use cw_storage_plus::MultiIndex;
+    /// use serde::{Deserialize, Serialize};
+    ///
+    /// #[derive(Deserialize, Serialize, Clone)]
+    /// struct Data {
+    ///     pub name: String,
+    ///     pub age: u32,
+    /// }
+    ///
+    /// MultiIndex::new(
+    ///     |d: &Data, k: Vec<u8>| (d.age, k),
+    ///     "age",
+    ///     "age__owner",
+    /// );
+    /// ```
     pub fn new(
         idx_fn: fn(&T, Vec<u8>) -> K,
         pk_namespace: &'a str,
@@ -77,8 +101,8 @@ where
 fn deserialize_multi_kv<T: DeserializeOwned>(
     store: &dyn Storage,
     pk_namespace: &[u8],
-    kv: Pair,
-) -> StdResult<Pair<T>> {
+    kv: Record,
+) -> StdResult<Record<T>> {
     let (key, pk_len) = kv;
 
     // Deserialize pk_len
@@ -120,7 +144,7 @@ where
     T: Serialize + DeserializeOwned + Clone,
     K: PrimaryKey<'a>,
 {
-    pub fn prefix(&self, p: K::Prefix) -> Prefix<T> {
+    pub fn prefix(&self, p: K::Prefix) -> Prefix<Vec<u8>, T> {
         Prefix::with_deserialization_function(
             self.idx_namespace,
             &p.prefix(),
@@ -129,7 +153,7 @@ where
         )
     }
 
-    pub fn sub_prefix(&self, p: K::SubPrefix) -> Prefix<T> {
+    pub fn sub_prefix(&self, p: K::SubPrefix) -> Prefix<Vec<u8>, T> {
         Prefix::with_deserialization_function(
             self.idx_namespace,
             &p.prefix(),
@@ -138,47 +162,37 @@ where
         )
     }
 
+    fn no_prefix(&self) -> Prefix<Vec<u8>, T> {
+        Prefix::with_deserialization_function(
+            self.idx_namespace,
+            &[],
+            self.pk_namespace,
+            deserialize_multi_kv,
+        )
+    }
+
     pub fn index_key(&self, k: K) -> Vec<u8> {
         k.joined_key()
     }
 
-    // FIXME?: Move to Prefix<T> for ergonomics
-    pub fn pks<'c>(
-        &self,
-        store: &'c dyn Storage,
-        p: K::Prefix,
-        min: Option<Bound>,
-        max: Option<Bound>,
-        order: Order,
-    ) -> Box<dyn Iterator<Item = StdResult<Vec<u8>>> + 'c>
-    where
-        T: 'c,
-    {
-        let prefix = self.prefix(p);
-        let mapped = prefix.range(store, min, max, order).map(|res| {
-            let t = res?;
-            Ok(t.0)
-        });
-        Box::new(mapped)
-    }
-
     #[cfg(test)]
     pub fn count(&self, store: &dyn Storage, p: K::Prefix) -> usize {
-        self.pks(store, p, None, None, Order::Ascending).count()
+        let prefix = self.prefix(p);
+        prefix.keys(store, None, None, Order::Ascending).count()
     }
 
     #[cfg(test)]
     pub fn all_pks(&self, store: &dyn Storage, p: K::Prefix) -> Vec<Vec<u8>> {
-        self.pks(store, p, None, None, Order::Ascending)
-            .collect::<StdResult<Vec<Vec<u8>>>>()
-            .unwrap()
+        let prefix = self.prefix(p);
+        prefix
+            .keys(store, None, None, Order::Ascending)
+            .collect::<Vec<Vec<u8>>>()
     }
 
     #[cfg(test)]
-    pub fn all_items(&self, store: &dyn Storage, prefix: K::Prefix) -> StdResult<Vec<Pair<T>>> {
-        self.prefix(prefix)
-            .range(store, None, None, Order::Ascending)
-            .collect()
+    pub fn all_items(&self, store: &dyn Storage, p: K::Prefix) -> StdResult<Vec<Record<T>>> {
+        let prefix = self.prefix(p);
+        prefix.range(store, None, None, Order::Ascending).collect()
     }
 }
 
@@ -187,7 +201,6 @@ impl<'a, K, T> MultiIndex<'a, K, T>
 where
     T: Serialize + DeserializeOwned + Clone,
     K: PrimaryKey<'a>,
-    K::SubPrefix: EmptyPrefix,
 {
     // I would prefer not to copy code from Prefix, but no other way
     // with lifetimes (create Prefix inside function and return ref = no no)
@@ -197,15 +210,45 @@ where
         min: Option<Bound>,
         max: Option<Bound>,
         order: Order,
-    ) -> Box<dyn Iterator<Item = StdResult<Pair<T>>> + 'c>
+    ) -> Box<dyn Iterator<Item = StdResult<Record<T>>> + 'c>
     where
         T: 'c,
     {
-        self.sub_prefix(K::SubPrefix::new())
-            .range(store, min, max, order)
+        self.no_prefix().range(store, min, max, order)
+    }
+
+    pub fn keys<'c>(
+        &'c self,
+        store: &'c dyn Storage,
+        min: Option<Bound>,
+        max: Option<Bound>,
+        order: Order,
+    ) -> Box<dyn Iterator<Item = Vec<u8>> + 'c> {
+        self.no_prefix().keys(store, min, max, order)
+    }
+
+    /// while range assumes you set the prefix to one element and call range over the last one,
+    /// prefix_range accepts bounds for the lowest and highest elements of the Prefix we wish to
+    /// accept, and iterates over those. There are some issues that distinguish these to and blindly
+    /// casting to Vec<u8> doesn't solve them.
+    pub fn prefix_range<'c>(
+        &'c self,
+        store: &'c dyn Storage,
+        min: Option<PrefixBound<'a, K::Prefix>>,
+        max: Option<PrefixBound<'a, K::Prefix>>,
+        order: cosmwasm_std::Order,
+    ) -> Box<dyn Iterator<Item = StdResult<cosmwasm_std::Record<T>>> + 'c>
+    where
+        T: 'c,
+        'a: 'c,
+    {
+        let mapped = namespaced_prefix_range(store, self.idx_namespace, min, max, order)
+            .map(move |kv| (deserialize_multi_kv)(store, self.pk_namespace, kv));
+        Box::new(mapped)
     }
 }
 
+/// UniqueRef stores Binary(Vec[u8]) representation of private key and index value
 #[derive(Deserialize, Serialize)]
 pub(crate) struct UniqueRef<T> {
     // note, we collapse the pk - combining everything under the namespace - even if it is composite
@@ -223,6 +266,23 @@ pub struct UniqueIndex<'a, K, T> {
 
 impl<'a, K, T> UniqueIndex<'a, K, T> {
     // TODO: make this a const fn
+    /// Create a new UniqueIndex
+    ///
+    /// idx_fn - lambda creating index key from index value
+    /// idx_namespace - prefix for the index value
+    ///
+    /// ## Example:
+    ///
+    /// ```rust
+    /// use cw_storage_plus::{U32Key, UniqueIndex};
+    ///
+    /// struct Data {
+    ///     pub name: String,
+    ///     pub age: u32,
+    /// }
+    ///
+    /// UniqueIndex::new(|d: &Data| U32Key::new(d.age), "data__age");
+    /// ```
     pub fn new(idx_fn: fn(&T) -> K, idx_namespace: &'a str) -> Self {
         UniqueIndex {
             index: idx_fn,
@@ -260,7 +320,7 @@ where
     }
 }
 
-fn deserialize_unique_kv<T: DeserializeOwned>(kv: Pair) -> StdResult<Pair<T>> {
+fn deserialize_unique_kv<T: DeserializeOwned>(kv: Record) -> StdResult<Record<T>> {
     let (_, v) = kv;
     let t = from_slice::<UniqueRef<T>>(&v)?;
     Ok((t.pk.into(), t.value))
@@ -275,20 +335,26 @@ where
         k.joined_key()
     }
 
-    pub fn prefix(&self, p: K::Prefix) -> Prefix<T> {
+    pub fn prefix(&self, p: K::Prefix) -> Prefix<Vec<u8>, T> {
         Prefix::with_deserialization_function(self.idx_namespace, &p.prefix(), &[], |_, _, kv| {
             deserialize_unique_kv(kv)
         })
     }
 
-    pub fn sub_prefix(&self, p: K::SubPrefix) -> Prefix<T> {
+    pub fn sub_prefix(&self, p: K::SubPrefix) -> Prefix<Vec<u8>, T> {
         Prefix::with_deserialization_function(self.idx_namespace, &p.prefix(), &[], |_, _, kv| {
             deserialize_unique_kv(kv)
         })
     }
 
+    fn no_prefix(&self) -> Prefix<Vec<u8>, T> {
+        Prefix::with_deserialization_function(self.idx_namespace, &[], &[], |_, _, kv| {
+            deserialize_unique_kv(kv)
+        })
+    }
+
     /// returns all items that match this secondary index, always by pk Ascending
-    pub fn item(&self, store: &dyn Storage, idx: K) -> StdResult<Option<Pair<T>>> {
+    pub fn item(&self, store: &dyn Storage, idx: K) -> StdResult<Option<Record<T>>> {
         let data = self
             .idx_map
             .may_load(store, idx)?
@@ -302,7 +368,6 @@ impl<'a, K, T> UniqueIndex<'a, K, T>
 where
     T: Serialize + DeserializeOwned + Clone,
     K: PrimaryKey<'a>,
-    K::SubPrefix: EmptyPrefix,
 {
     // I would prefer not to copy code from Prefix, but no other way
     // with lifetimes (create Prefix inside function and return ref = no no)
@@ -312,11 +377,20 @@ where
         min: Option<Bound>,
         max: Option<Bound>,
         order: Order,
-    ) -> Box<dyn Iterator<Item = StdResult<Pair<T>>> + 'c>
+    ) -> Box<dyn Iterator<Item = StdResult<Record<T>>> + 'c>
     where
         T: 'c,
     {
-        self.sub_prefix(K::SubPrefix::new())
-            .range(store, min, max, order)
+        self.no_prefix().range(store, min, max, order)
+    }
+
+    pub fn keys<'c>(
+        &self,
+        store: &'c dyn Storage,
+        min: Option<Bound>,
+        max: Option<Bound>,
+        order: Order,
+    ) -> Box<dyn Iterator<Item = Vec<u8>> + 'c> {
+        self.no_prefix().keys(store, min, max, order)
     }
 }
