@@ -1,14 +1,40 @@
-use cosmwasm_std::Addr;
+use cosmwasm_std::{Addr, Timestamp};
 use std::marker::PhantomData;
 
+use crate::de::KeyDeserialize;
 use crate::helpers::namespaces_with_key;
 use crate::Endian;
 
-// pub trait PrimaryKey<'a>: Copy {
+/// `PrimaryKey` needs to be implemented for types that want to be a `Map` (or `Map`-like) key,
+/// or part of a key.
+///
+/// In particular, it defines a series of types that help iterating over parts of a (composite) key:
+///
+/// `Prefix`: Prefix is eager. That is, except for empty keys, it's always "one less" than the full key.
+/// `Suffix`: Suffix is the complement of prefix.
+/// `SubPrefix`: Sub-prefix is "one less" than prefix.
+/// `SuperSuffix`: Super-suffix is "one more" than suffix. The complement of sub-prefix.
+///
+/// By example, for a 2-tuple `(T, U)`:
+///
+/// `T`: Prefix.
+/// `U`: Suffix.
+/// `()`: Sub-prefix.
+/// `(T, U)`: Super-suffix.
+///
+/// `SubPrefix` and `SuperSuffix` only make real sense in the case of triples. Still, they need to be
+/// consistently defined for all types.
 pub trait PrimaryKey<'a>: Clone {
+    /// These associated types need to implement `Prefixer`, so that they can be useful arguments
+    /// for `prefix()`, `sub_prefix()`, and their key-deserializable variants.
     type Prefix: Prefixer<'a>;
     type SubPrefix: Prefixer<'a>;
 
+    /// These associated types need to implement `KeyDeserialize`, so that they can be returned from
+    /// `range_de()` and friends.
+    type Suffix: KeyDeserialize;
+    type SuperSuffix: KeyDeserialize;
+
     /// returns a slice of key steps, which can be optionally combined
     fn key(&self) -> Vec<&[u8]>;
 
@@ -21,8 +47,10 @@ pub trait PrimaryKey<'a>: Clone {
 
 // Empty / no primary key
 impl<'a> PrimaryKey<'a> for () {
-    type Prefix = ();
-    type SubPrefix = ();
+    type Prefix = Self;
+    type SubPrefix = Self;
+    type Suffix = Self;
+    type SuperSuffix = Self;
 
     fn key(&self) -> Vec<&[u8]> {
         vec![]
@@ -32,6 +60,8 @@ impl<'a> PrimaryKey<'a> for () {
 impl<'a> PrimaryKey<'a> for &'a [u8] {
     type Prefix = ();
     type SubPrefix = ();
+    type Suffix = Self;
+    type SuperSuffix = Self;
 
     fn key(&self) -> Vec<&[u8]> {
         // this is simple, we don't add more prefixes
@@ -43,6 +73,8 @@ impl<'a> PrimaryKey<'a> for &'a [u8] {
 impl<'a> PrimaryKey<'a> for &'a str {
     type Prefix = ();
     type SubPrefix = ();
+    type Suffix = Self;
+    type SuperSuffix = Self;
 
     fn key(&self) -> Vec<&[u8]> {
         // this is simple, we don't add more prefixes
@@ -51,9 +83,13 @@ impl<'a> PrimaryKey<'a> for &'a str {
 }
 
 // use generics for combining there - so we can use &[u8], Vec<u8>, or IntKey
-impl<'a, T: PrimaryKey<'a> + Prefixer<'a>, U: PrimaryKey<'a>> PrimaryKey<'a> for (T, U) {
+impl<'a, T: PrimaryKey<'a> + Prefixer<'a> + KeyDeserialize, U: PrimaryKey<'a> + KeyDeserialize>
+    PrimaryKey<'a> for (T, U)
+{
     type Prefix = T;
     type SubPrefix = ();
+    type Suffix = U;
+    type SuperSuffix = Self;
 
     fn key(&self) -> Vec<&[u8]> {
         let mut keys = self.0.key();
@@ -63,11 +99,17 @@ impl<'a, T: PrimaryKey<'a> + Prefixer<'a>, U: PrimaryKey<'a>> PrimaryKey<'a> for
 }
 
 // use generics for combining there - so we can use &[u8], Vec<u8>, or IntKey
-impl<'a, T: PrimaryKey<'a> + Prefixer<'a>, U: PrimaryKey<'a> + Prefixer<'a>, V: PrimaryKey<'a>>
-    PrimaryKey<'a> for (T, U, V)
+impl<
+        'a,
+        T: PrimaryKey<'a> + Prefixer<'a>,
+        U: PrimaryKey<'a> + Prefixer<'a> + KeyDeserialize,
+        V: PrimaryKey<'a> + KeyDeserialize,
+    > PrimaryKey<'a> for (T, U, V)
 {
     type Prefix = (T, U);
     type SubPrefix = T;
+    type Suffix = V;
+    type SuperSuffix = (U, V);
 
     fn key(&self) -> Vec<&[u8]> {
         let mut keys = self.0.key();
@@ -77,10 +119,14 @@ impl<'a, T: PrimaryKey<'a> + Prefixer<'a>, U: PrimaryKey<'a> + Prefixer<'a>, V:
     }
 }
 
-// pub trait Prefixer<'a>: Copy {
 pub trait Prefixer<'a> {
-    /// returns 0 or more namespaces that should length-prefixed and concatenated for range searches
+    /// returns 0 or more namespaces that should be length-prefixed and concatenated for range searches
     fn prefix(&self) -> Vec<&[u8]>;
+
+    fn joined_prefix(&self) -> Vec<u8> {
+        let prefixes = self.prefix();
+        namespaces_with_key(&prefixes, &[])
+    }
 }
 
 impl<'a> Prefixer<'a> for () {
@@ -119,18 +165,11 @@ impl<'a> Prefixer<'a> for &'a str {
     }
 }
 
-// this is a marker for the Map.range() helper, so we can detect () in Generic bounds
-pub trait EmptyPrefix {
-    fn new() -> Self;
-}
-
-impl EmptyPrefix for () {
-    fn new() {}
-}
-
 impl<'a> PrimaryKey<'a> for Vec<u8> {
     type Prefix = ();
     type SubPrefix = ();
+    type Suffix = Self;
+    type SuperSuffix = Self;
 
     fn key(&self) -> Vec<&[u8]> {
         vec![&self]
@@ -146,6 +185,8 @@ impl<'a> Prefixer<'a> for Vec<u8> {
 impl<'a> PrimaryKey<'a> for String {
     type Prefix = ();
     type SubPrefix = ();
+    type Suffix = Self;
+    type SuperSuffix = Self;
 
     fn key(&self) -> Vec<&[u8]> {
         vec![self.as_bytes()]
@@ -162,6 +203,8 @@ impl<'a> Prefixer<'a> for String {
 impl<'a> PrimaryKey<'a> for &'a Addr {
     type Prefix = ();
     type SubPrefix = ();
+    type Suffix = Self;
+    type SuperSuffix = Self;
 
     fn key(&self) -> Vec<&[u8]> {
         // this is simple, we don't add more prefixes
@@ -179,6 +222,8 @@ impl<'a> Prefixer<'a> for &'a Addr {
 impl<'a> PrimaryKey<'a> for Addr {
     type Prefix = ();
     type SubPrefix = ();
+    type Suffix = Self;
+    type SuperSuffix = Self;
 
     fn key(&self) -> Vec<&[u8]> {
         // this is simple, we don't add more prefixes
@@ -193,9 +238,14 @@ impl<'a> Prefixer<'a> for Addr {
 }
 
 // this auto-implements PrimaryKey for all the IntKey types
-impl<'a, T: Endian + Clone> PrimaryKey<'a> for IntKey<T> {
+impl<'a, T: Endian + Clone> PrimaryKey<'a> for IntKey<T>
+where
+    IntKey<T>: KeyDeserialize,
+{
     type Prefix = ();
     type SubPrefix = ();
+    type Suffix = Self;
+    type SuperSuffix = Self;
 
     fn key(&self) -> Vec<&[u8]> {
         self.wrapped.key()
@@ -250,7 +300,8 @@ impl<T: Endian> From<T> for IntKey<T> {
 
 impl<T: Endian> From<Vec<u8>> for IntKey<T> {
     fn from(wrap: Vec<u8>) -> Self {
-        // TODO: assert proper length
+        // TODO: Consider properly handling case, when `wrap` has length not conforming for the
+        // wrapped integer type.
         IntKey {
             wrapped: wrap,
             data: PhantomData,
@@ -264,6 +315,44 @@ impl<T: Endian> From<IntKey<T>> for Vec<u8> {
     }
 }
 
+#[derive(Clone, Debug, PartialEq, Eq)]
+pub struct TimestampKey(U64Key);
+
+impl TimestampKey {
+    pub fn new(ts: Timestamp) -> Self {
+        Self(ts.nanos().into())
+    }
+}
+
+impl<'a> PrimaryKey<'a> for TimestampKey {
+    type Prefix = ();
+    type SubPrefix = ();
+    type Suffix = Self;
+    type SuperSuffix = Self;
+
+    fn key(&self) -> Vec<&[u8]> {
+        self.0.key()
+    }
+}
+
+impl<'a> Prefixer<'a> for TimestampKey {
+    fn prefix(&self) -> Vec<&[u8]> {
+        self.0.key()
+    }
+}
+
+impl From<Vec<u8>> for TimestampKey {
+    fn from(val: Vec<u8>) -> Self {
+        Self(val.into())
+    }
+}
+
+impl From<Timestamp> for TimestampKey {
+    fn from(val: Timestamp) -> Self {
+        Self::new(val)
+    }
+}
+
 #[cfg(test)]
 mod test {
     use super::*;
